; -----------------------------------------------------------------------------
; tn15-Bewegungsmelder-05.asm
;
; Misst die Spannung des Bewegungsmelders an ADC1 (PortB.2 = Pin 7)
; Die Referenzspannung ist VCC, Ruheausgang am Sensor ist VCC / 2.
;
; Ein Alarm wird nur ausgelst, wenn innerhalb von 'k_Mess_dauer'
; mehrfach der Ruhelevel des Sensors verlassen wird.
; Jede berschreitung verlngert die Messdauer um 'k-add2Mess_dauer', dadurch
; verlngert sich die Messdauer, je mehr berschreitungen sich ereignen,
; die Chance auf Alarmauslsung steigt
;
; Dabei wird zwischen geringfgigen berschreitungen (level1 = schwache Signale)
; und berschreitung im Bereich von GND bzw. VCC (level2 = starke Signale) unterschieden:
;
; level1: 	10x berschreitung low + high (kein Pegelwechsel notwendig)
; 			2x low + 2x high  (Pegelwechsel notwendig)
; level2:	5x berschreitung low + high (kein Pegelwechsel notwendig)
;   		1x low + 1x high  (Pegelwechsel notwendig)
;
; Sobald das Sensorausgang ausserhalb von Level1, wird
; - wenn noch kein Alarm ausgelst ist, wird Zhler Mess_dauer um
;  'k_add2Mess_Messdauer' erhht,
; - wenn bereits Alarm, wird Zhler Mess_dauer um 'k_Alarm_dauer'
;   zurckgesetzt,
;
; Wenn der Sensorausgang im Normalbereich und ein Alarm ausgelst, wird die Alarm_dauer decrementiert.
; Beim Auslsen des Alarms werden alle Zhler gelscht.
; Danach wird ausgewertet, wieviele Uberschreitungen vorlagen und ob ggf. Alarm
; ausgelst werden soll (siehe oben).
; Wenn Alarm ausgelst ist, wird getestet, ob die Voraussetzung fr Lschung vorhanden sind.
; Ein eigenstndiger Zhler schaltet einen externen Alarm, er wird bei jeder
; Alarmauslsung auf seinen Ausgangswert 'Alarm_out_dauer' zurckgesetzt.
;
.include "tn15def.inc"	; 1,6 MHz

.equ	k_preload		= 100 ; 256-156 = preload fr Timer
; 1 / 1.600.000 * 1024 * 156 = 0.1 Sek -> 10 Messungen/sec
; Zeitangaben sind bezogen auf die Anzahl der Messung / Sekunde

.equ	k_level1_high 	= 127 + 9	; Grenzwerte fr Ausgangsspannung des
.equ	k_level1_low	= 127 - 11  ; Bewegungsmelders
.equ	k_level2_high	= 255 - 60	; ca. 4.0V
.equ	k_level2_low	= 0 + 32		; ca. 0.6V
.equ	k_on				= 0			; LED's sind bei low eingeschaltet ...
.equ  k_off				= 1 	 		; LED's sind bei high ausgeschaltet ...

.equ	k_Mess_dauer 	= 30		; Zeitdauer, in der Messwertberschreitung kumuliert werden
										; werden die Werte von k_Cnt_levelx berschritten, wird Alarm ausgelst
.equ	k_add2Mess_dauer= 5		; Verlngerung der Mess_dauer bei berschreiten des Grenzwertes
.equ	k_Alarm_dauer	= 30		; Zeit ohne Grenzwertberschreitung, bis Alarm wieder aufgehoben wird
.equ	k_Cnt_level1	= 10		; Anzahl der berschreitungen fr Level1 (Summe high + low)
.equ	k_Cnt_level2	= 3 		; Anzahl der berschreitungen fr Level2 (Summe high + low)
.equ	k_Cnt_hi_and_lo= 2		; Anzahl der berschreitungen fr Level1 (sowohl high als auch low)
.equ	k_Alarm_out_dauer	= 30	; Schaltdauer fr externe Alarmzeit
.equ	k_Zeit_faktor	= 60		; Faktor fr externe Alarmzeit, 
										; Verzgerung = k_Alarm_out_dauer * k_Zeit_Faktor * Sekunden

.def	S					= R1	; Zwischenspeicher fr SREG
.def	Preload			= R2	; Hilfsspeicher fr Preloadwert
.def	Time_Sekunde	= R3	; Zhler fr abgelaufenen Sekunden
.def	Time_10tel		= R4	; Zhler fr die 10tel Sekunden
.def	Mess_dauer		= R5	; Zhler fr Mess_dauer
.def	level1_high		= R6  ; Zhler level1 berschreitung
.def	level1_low		= R7	; Zhler level1 Unterschreitung
.def	level2_high		= R8	; Zhler level2 berschreitung
.def	level2_low		= R9	; Zhler level2 Unterschreitung
.def	Null				= R10	; Hilfsregister fr Konstante 0x00
.def	Alarm_out		= R11	;

.def	temp				= R16
.def	ADC_in			= R17
.def	Alarm_dauer		= R18
.def	Flag				= R19
.def	temp1				= R20
.def	temp_isr			= R21	; Temp-Variable fr Timer0_isr

; -----------------------------------------------------------------------------
;
.cseg
.org 	0
	rjmp	Init		; RESET
	reti				; INT0 		External Interrupt Request 0
	reti				; PCINT		Pin Change Interrupt
	reti				; TIMER1 COMPA	Timer/Counter1 Compare Match A
	reti				; TIMER1 OVF	Timer/Counter1 Overflow
	rjmp	TIMER0_OVF	; TIMER0 OVF	Timer/Counter0 Overflow
	reti				; EE READY		Eeprom Ready
	reti				; ANALOG COMP	Analog Comparator
	reti				; ADC			ADC Conversion Complete

; -----------------------------------------------------------------------------
Init:
									; TIMER
	LDI	temp, 103			; kalibrieren auf 1.6MHz
	OUT	OSCCAL, temp

;			  76543210	  	; PORTS
	LDI	temp, 0b00011011		; PortB.2 = ADC1 als Eingang, Rest Ausgang
	OUT	DDRB, temp

	LDI	temp, 0b00011011		; alle LEDs aus
	OUT	PortB, temp
						; ADC Control und Status Register
;	LDI	temp, 0b10001101		; ADC enable, ADC Interrupt enable, Prescale 32
	LDI	temp, 1<<ADEN | 1<<ADIE | 1<<ADPS2 | 1<<ADPS0
	OUT	ADCSR, temp
						; ADC Multiplexer Selection Register
;	LDI	temp, 0b00100001	; Referenz = Vcc, ADLAR = 1, ADMUX=1
	LDI	temp, 1<<ADLAR |1<<MUX0
	OUT	ADMUX, temp
						; MCU Control Register
;	LDI	temp, 0b00101000	; Sleep enable, ADC-Noice Reduction mode
	LDI	temp, 1<<SE | 1<<SM0
	OUT	MCUCR, temp

	LDI	temp, 0			; Konstanten 0 in Hilfsregister 
	MOV	Null, temp

	LDI	temp, k_preload		; Preload-Wert fr TIMER0 in Hilfsregister
	MOV	Preload, temp
	
	SEI					; Interrupts aktivieren

	; nach dem Einschalten 'klebt' der Sensorausgang fr einige Zeit an VCC
	; und 'schwingt' dann noch eine Zeit lang zur anderen Seite aus

	;LDI	temp, 0b00010000		; alle drei grnen LEDs einschalten, rote aus
	LDI	temp, 0<<4 | k_on<<3 | k_on<<1 | k_on
	OUT	PortB, temp
	LDI	temp1, 15			; Zhler fr nachfolgende Warmlaufschleife

Warmlaufen:
 	rcall	wait1s

 	RCALL	Messung
	CPI	ADC_in, k_Level1_low
	BRLO	Warmlaufen			; solange kleiner als Level1_low zurck

	CPI	ADC_in, k_Level1_high
	BRSH	Warmlaufen			; solange grer als Level1_high zurck
						; nun liegt der Ausgang im neutralen Bereich
	;LDI	temp, 0b00010010		; uere beiden LEDs einschalten, mittlere aus
	LDI 	temp, 0<<4 | k_on<<3 | k_off<<1 | k_on
	OUT	PortB, temp
		
	dec 	temp1				; zustzliche Warteschleife aus Ruhepegel
	BRNE	Warmlaufen			; damit beim Einschalten kein Alarm ausgelst wird
						
						; Hier ist endgltig Schluss mit Warten ....
	LDI	temp, 1<<TOIE0		; Timer0 overflow enable
	OUT	TIMSK, temp

	; 1 / 1.600.000 * 1024 = 0.64ms * 156 = 0.1sec
	LDI	temp, 1<<CS02 | 1<<CS00; Prescale = 1024
	OUT	TCCR0, temp			; Timer starten, zunchst ohne Preload
	
;	rcall	wait1s			; nach Aktivierung des Timers noch mal warten

	MOV	Alarm_out, Null		; alle Zhler auf Ausgangsposition
	MOV	Mess_dauer, null
	MOV	Flag, null
	MOV	level1_low,  Null		; alle Variablen zurcksetzen
	MOV	level1_high, Null
	MOV	level2_low,  Null
	MOV	level2_high, Null


; die Arbeit kann beginnen ----------------------------------------------------
Main:
	; Flag.0 wird von TIMER0_ISR alle 0.1s auf 1 gesetzt,
	; und am Ende von Main wieder gelscht
	SBRS	Flag, 0			; Flag.0 testen
	rjmp	main_exit			; wenn Flag.0 = 0, dann alles berspringen

						; wenn Flag.0 = 1, dann hier weiterarbeiten
	rcall Messung			; Messung ausfhren
	rcall	Signaltest

	SBRC	Flag, 7			; Flag.7 testen, 1 = Alarm, 0 = kein Alarm
	rjmp	main1				; wenn Flag.7 = 1, dann springen

	rcall	Test_auf_Alarm		; ausfhren, wenn Flag.7 = 0 -> (noch) kein Alarm

Main1:
	SBRC	Flag, 7			; Flag.7 testen, 1 = Alarm intern, 0 = kein Alarm
	rcall	Alarm_intern_aus		; wenn Flag.7 = 1, prfen, ob Alarmende erreicht

	SBRC	Flag, 6			; Flag.6 testen, 1 = Alarm extern, 0 = kein Alarm
	rcall	Alarm_extern_aus		; wenn Flag.6 = 1, prfen, ob Alarmende erreicht ist

	CBR	Flag, 0b00000001		; Flag.0 lschen, nur ein Durchgang je Timer0_ISR

Main_exit:					; hier gehts weiter, wenn Flag.0 = 0
	nop

	rjmp	main				; Endlosschleifen

; SUBROUTINEN -----------------------------------------------------------------
;
; Fhrte eine Spannungsmessung am Ausgang des Sensors durch,
; vergleicht den Wert mit 4 Schwellwerten level1/2 high/low,
; bei berschreitung werden die zustndigen Zhler incrementiert.
;
Messung:					; zuerst die Messung durchfhren
	SBI	ADCSR, ADSC			; Start Conversion
	sleep					; ISR fhrt nur 'reti' aus und startet das Programm
	nop

;	IN	ADC_in, ADCL		; zuerst ADCL einlesen (berflssig ??)
	IN	ADC_in, ADCH		; dann ADCH ins selbe Register, da nur 8Bit ausgewertet

	ret
; -----------------------------------------------------------------------------
;
Signaltest:
	CPI	ADC_in, k_level1_low	; nun mit den Grenzwerten vergleichen
	BRLO	lev1_low			; wenn kleiner, dann verzweigen

	CPI	ADC_in, k_level1_high
	BRSH	lev1_high			; wenn grer, dann verzweigen

lev_normal:					; hier liegt die Messung im Ruhebereich
	;LDI	temp, 0b00001011		; Muster fr LED 000
	LDI	temp, k_off<<3 | k_off<<1 | k_off

	SBRS	Flag, 7			; Alarmflag testen
	rjmp	Messung_exit		; wenn nicht gesetzt, dann zum Ende springen
						; wenn gesetzt, dann
	tst	Mess_dauer			; vor Decrementieren auf Null testen
	BREQ	Messung_exit		; wenn schon Null, dann is' gut

	dec	Mess_dauer			; den Zhler fr Mess_dauer decr.
						; das Unterprogramm alarm_intern_aus prft auf 0
	rjmp 	Messung_exit

lev1_low:					; der Messwert ist kleiner als k_level1_low
	CPI	ADC_in, k_level2_low
	BRLO	lev2_low			; wenn kleiner, dann springen
	inc	level1_low			; Zhler fr Unterschreitung
	;LDI	temp, 0b00001010		; Muster fr LED 00X
	LDI	temp, k_off<<3 | k_off<<1 | k_on
	rjmp	Messung_exit_alarm

lev2_low:					; der Messwert ist kleiner als k_level2_low
	inc	level2_low
	;LDI	temp, 0b00001000		; Muster fr LED 0XX
	LDI	temp, k_off<<3 | k_on<<1 | k_on
	rjmp	Messung_exit_alarm

lev1_high:					; der Messwert ist grer als k_level1_high
	CPI	ADC_in, k_level2_high
	BRSH	lev2_high			; wenn grer springen
						; wenn kleiner als k_level2_high
	inc	level1_high			; Zhler fr berschreitung
	;LDI	temp, 0b00000011		; Muster fr LED X00
	LDI	temp, k_on<<3 | k_off<<1 | k_off
	rjmp	Messung_exit_alarm

lev2_high:					; der Messwert ist grer als k_level2_high
	inc	level2_high			; Zhler incr.
	;LDI	temp, 0b00000001		; Muster fr LED XX0
	LDI	temp, k_on<<3 | k_on<<1 | k_off
;	rjmp	Messung_exit_alarm

Messung_exit_alarm:			; der Messwert lag ausserhalb des Normallevel,
	SBRS	Flag, 7			; Flag.7 testen, 1 = Alarm besteht, 0 = kein Alarm
						; wenn kein Alarm, dann die Mess_dauer verlngern
						; wenn Alarm besteht, dann Alarm_dauer zurcksetzen
	rjmp	Messung_exit_1		; wenn 0, dann verzweigen
						; wenn 1, dann besteht bereits ein Alarm
						; dauer verlngern
	LDI	temp1, k_Alarm_dauer	; die Alarm_dauer auf grten Wert
	MOV	Mess_dauer, temp1		; und in Zhler schreiben

	LDI	temp1, k_Alarm_out_dauer
	MOV	Alarm_out, temp1		; auch externe Alarmzeit zurcksetzen
	rjmp	Messung_exit

Messung_exit_1:				; noch kein Alarm ausgelst
						; dann die Mess_dauer verlngern
	LDI	Temp1, k_add2Mess_dauer
	ADD	Mess_dauer, temp1		; Zhler fr die Mess_dauer erhhen

Messung_exit:
	IN	temp1, PortB		; Status von PortB einlesen
	ANDI	temp1, 0B00010100		; PortB.4 (Alarmstatus) und 2 (Input) beibehalten
	OR	temp1, temp			; LEDs hineinkopieren
	OUT	PortB, temp1		; und auf PortB rausschreiben

	ret
; -----------------------------------------------------------------------------
;
; Testet, ob Alarmbedingungen erfllt sind,
; dazu werden die Zhlerstnde der Schwellwertberschreitungen ausgewertet.
; zuerst wird berprft, ob die Summe der berschreitung von Level1/2 Grenzwerte berschreitet,
; danach wird berprft, ob Einzelwerte ihre Grenzwerte von Level1 berschreiten
Test_auf_Alarm:
	; testen, ob Anzahl der High-Alarme	berschritten
	MOV	temp, level2_high		; Summe high + low Alarme von Level2 addieren
	ADD	temp, level2_low

	CPI	temp, k_Cnt_Level2	; wenn => dann ist Alarm fllig
	BRSH	Alarm_Set
						; wenn nicht, dann nchsten Test
						; Pegelwechsel bei High-alarm
	tst	level2_high
	BREQ	Test_auf_Alarm_2		; wenn 0, dann kein Alarm, weiter testen

	tst	level2_low
	BREQ	Test_auf_Alarm_2		; wenn 0, dann kein Alarm, weiter testen
	rjmp	Alarm_Set			; beide High-Pegel sind > 0

Test_auf_Alarm_2:
	MOV	temp, level1_high		; Summe high + low Alarme von Level1 addieren
	ADD	temp, level1_low
	CPI	temp, k_Cnt_Level1
	BRSH	Alarm_Set  			; wenn count_Level1 erreicht, dann Alarm
						
						; Wenn ein Pegelwechsel stattgefunden hat, dann
						; mssen LOW und HIGH Grenzwert berschritten haben 
	LDI	temp, k_Cnt_hi_and_lo	; den Grenzwert in temp einlesen
	CP	level1_high, temp		; mit High vergleichen
	BRLO	Test_auf_Alarm_Exit	; wenn < temp, dann kein Alarm
						; High hat die notwendigen berschreitungen erfllt
	CP	level1_low, temp		; Low vergleichen
	BRSH	Alarm_Set 			; wenn => temp, dann ist Alarm fllig

	rjmp	Test_auf_Alarm_exit	; wenn < temp, dann kein Alarm

Alarm_Set:					; die Alarmbedingung ist erfllt ...
	SBR	Flag, 0b11000000		; Flag.7 und Flag.6 fr Alarm und Alarm_dauer setzen

	LDI	temp, k_Alarm_dauer 	; Die Lnge des internen Alarms laden
	MOV	Mess_dauer, temp		; und in Zhler schreiben

	LDI	temp, k_Alarm_out_dauer ; Die Lnge des externen Alarms laden
	MOV	Alarm_out, temp		; und in Zhler schreiben

	LDI	temp, 10			; Zhler fr 10tel Sekunde vorbelegen
	MOV	Time_10tel, temp

	LDI	temp, k_Zeit_Faktor	; Zhler fr Sekunden vorbelegen
	MOV	Time_Sekunde, temp

	MOV	level1_low,  Null		; alle Variablen zurcksetzen
	MOV	level1_high, Null
	MOV	level2_low,  Null
	MOV	level2_high, Null

	OUT	TCNT0, Preload		; TimerCounter zurcksetzen

Test_auf_Alarm_Exit:
	ret

; -----------------------------------------------------------------------------
; Testet, ob Alarmzeit intern abgelaufen ist
Alarm_intern_Aus:
	TST 	Mess_dauer			; test, ob Mess_dauer = 0
	BRNE	Alarm_intern_exit		; wenn grer, dann beenden
						; wenn 0, dann
	CBR	Flag, 0b10000000		; Flag.7 zurcksetzen

Alarm_intern_exit:
	ret

; -----------------------------------------------------------------------------
; Testet, ob Alarmzeit extern abgelaufen ist
Alarm_extern_Aus:
	TST 	Alarm_out			; test, ob Alarm_dauer = 0
	BRNE	Alarm_extern_exit		; wenn grer, dann beenden
						; wenn 0, dann
	CBR	Flag, 0b01000000		; Flag.6 zurcksetzen
	
Alarm_extern_exit:
	ret

; -----------------------------------------------------------------------------
TIMER0_OVF:
	IN	S, SREG			; SREG sichern
	OUT	TCNT0, preload		; Preload-Wert fr TCNT0 laden

	SBRS	Flag, 6			; Alarm_out_flag testen
	rjmp	TIMER0_OVF_exit		; springen, wenn Flag.6 = 0

						; hier gehts weiter, wenn Flag.6 = 1
	dec	Time_10tel			; Zhler fr 10tel Sekunde decr.
	BRNE	TIMER0_OVF_exit		; wenn nicht Null, dann beenden

						; wenn 10tel-Sekunde = Null, dann
	LDI	temp_isr, 10		; 10 in den Zhler der 10tel Sekunde laden
	MOV	Time_10tel, temp_isr

	dec	Time_Sekunde		; Zhler fr Sekunden decrementieren
	BRNE	TIMER0_OVF_exit		; wenn Sekunden > 0, dann beenden

						; wenn Sekunden = Null, dann
	LDI	temp_isr, k_Zeit_Faktor	; neuen Startwert in Sekunde schreiben
	MOV	Time_Sekunde, temp_isr

	dec	Alarm_out			; wird durch alarm_extern_aus geprft

TIMER0_OVF_exit:
	MOV	temp_isr, Flag      	; Zustand von Flag.x ausgeben, dazu Flag kopieren
	
	;com	temp_isr			; invertieren, da bei 0 die LED leuchtet ...
	
	BST	temp_isr,6			; Bit.x nach T
	in	temp_isr, PortB		; PortB laden
	BLD	temp_isr, 4			; T in PortB.4 kopieren
	OUT	PortB, temp_isr		; und PortB schreiben

	ORI	Flag, 0b00000001		; Flag.0 setzen zum Auslsen einer neuen Messung

	OUT	SREG, S			; SREG restaurieren
	reti

; -----------------------------------------------------------------------------
wait1s:

; Warteschleifen-Generator - warte 1599999 Zyklen
          ldi  R29, $F1
WGLOOP0:  ldi  R30, $0E
WGLOOP1:  ldi  R31, $9D
WGLOOP2:  dec  R31
          brne WGLOOP2
          dec  R30
          brne WGLOOP1
          dec  R29
          brne WGLOOP0
	    ret
.exit	; end of file -----------------------------------------------------------
